#
# Copyright (c) 2016 NEC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re
import select
import time
import traceback

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder import utils

LOG = logging.getLogger(__name__)

retry_msgids = ['iSM31005', 'iSM31015', 'iSM42408', 'iSM42412', 'iSM19411']


def get_sleep_time_for_clone(retry_count):
    if retry_count < 19:
        return int(10.0 * (1.1 ** retry_count))
    else:
        return 60


class MStorageISMCLI(object):
    """SSH client."""

    def __init__(self, properties):
        super(MStorageISMCLI, self).__init__()

        self._sshpool = None
        self._properties = properties

    def _execute(self, command, expected_status=[0], raise_exec=True):
        return self._sync_execute(command, self._properties['diskarray_name'],
                                  expected_status, raise_exec)

    @coordination.synchronized('mstorage_ismcli_execute_{diskarray_name}')
    def _sync_execute(self, command, diskarray_name,
                      expected_status=[0], raise_exec=True):
        retry_flag = True
        retry_count = 0
        while retry_flag is True:
            try:
                out, err, status = self._cli_execute(command, expected_status,
                                                     False)
                if status != 0:
                    errflg = 0
                    errnum = out + err
                    LOG.debug('ismcli failed (errnum=%s).', errnum)
                    for retry_msgid in retry_msgids:
                        if errnum.find(retry_msgid) >= 0:
                            LOG.debug('`%(command)s` failed. '
                                      '%(name)s %(errnum)s '
                                      'retry_count=%(retry_count)d',
                                      {'command': command,
                                       'name': __name__,
                                       'errnum': errnum,
                                       'retry_count': retry_count})
                            errflg = 1
                            break
                    if errflg == 1:
                        retry_count += 1
                        if retry_count >= 60:
                            msg = (_('Timeout `%(command)s`.'
                                     ' status=%(status)d, '
                                     'out="%(out)s", '
                                     'err="%(err)s".') %
                                   {'command': command,
                                    'status': status,
                                    'out': out,
                                    'err': err})
                            raise exception.APITimeout(msg)
                        time.sleep(5)
                        continue
                    else:
                        if raise_exec is True:
                            msg = _('Command `%s` failed.') % command
                            raise exception.VolumeBackendAPIException(data=msg)
            except EOFError:
                with excutils.save_and_reraise_exception() as ctxt:
                    LOG.debug('EOFError has occurred. '
                              '%(name)s retry_count=%(retry_count)d',
                              {'name': __name__,
                               'retry_count': retry_count})
                    retry_count += 1
                    if retry_count < 60:
                        ctxt.reraise = False
                time.sleep(5)
                continue
            retry_flag = False

        return out, err, status

    def _execute_nolock(self, command, expected_status=[0], raise_exec=True):
        retry_flag = True
        retry_count = 0
        while retry_flag is True:
            try:
                out, err, status = self._cli_execute(command, expected_status,
                                                     raise_exec)
            except EOFError:
                with excutils.save_and_reraise_exception() as ctxt:
                    LOG.debug('EOFError has occurred. '
                              '%(name)s retry_count=%(retry_count)d',
                              {'name': __name__,
                               'retry_count': retry_count})
                    retry_count += 1
                    if retry_count < 60:
                        ctxt.reraise = False
                time.sleep(5)
                continue
            retry_flag = False
        return out, err, status

    def _cli_execute(self, command, expected_status=[0], raise_exec=True):
        if not self._sshpool:
            LOG.debug('ssh_utils.SSHPool execute.')
            self._sshpool = ssh_utils.SSHPool(
                self._properties['cli_fip'],
                self._properties['ssh_pool_port_number'],
                self._properties['ssh_conn_timeout'],
                self._properties['cli_user'],
                self._properties['cli_password'],
                privatekey=self._properties['cli_privkey'])

        with self._sshpool.item() as ssh:
            LOG.debug('`%s` executing...', command)
            stdin, stdout, stderr = ssh.exec_command(command)
            stdin.close()
            channel = stdout.channel
            tmpout, tmperr = b'', b''
            while 1:
                select.select([channel], [], [])
                if channel.recv_ready():
                    tmpout += channel.recv(4096)
                    continue
                if channel.recv_stderr_ready():
                    tmperr += channel.recv_stderr(4096)
                    continue
                if channel.exit_status_ready():
                    status = channel.recv_exit_status()
                    break
            LOG.debug('`%(command)s` done. status=%(status)d.',
                      {'command': command, 'status': status})
            out = utils.convert_str(tmpout)
            err = utils.convert_str(tmperr)
            if expected_status is not None and status not in expected_status:
                LOG.debug('`%(command)s` failed. status=%(status)d, '
                          'out="%(out)s", err="%(err)s".',
                          {'command': command, 'status': status,
                           'out': out, 'err': err})
                if raise_exec is True:
                    msg = _('Command `%s` failed.') % command
                    raise exception.VolumeBackendAPIException(data=msg)
            return out, err, status

    def view_all(self, conf_ismview_path=None, delete_ismview=True,
                 cmd_lock=True):
        if self._properties['queryconfig_view'] is True:
            command = 'clioutmsg xml; iSMview'
            if self._properties['ismview_alloptimize'] is True:
                command += ' --alloptimize'
            else:
                command += ' -all'
        else:
            command = 'iSMquery -cinder -xml -all'
        if cmd_lock is True:
            out, err, status = self._execute(command)
        else:
            out, err, status = self._execute_nolock(command)

        exstats = re.compile(r"(.*)ExitStatus(.*)\n")
        tmpout = exstats.sub('', out)
        out = tmpout
        if conf_ismview_path is not None:
            if delete_ismview:
                if os.path.exists(conf_ismview_path):
                    os.remove(conf_ismview_path)
                    LOG.debug('Remove clioutmsg xml to %s.',
                              conf_ismview_path)
            else:
                with open(conf_ismview_path, 'w+') as f:
                    f.write(out)
                    LOG.debug('Wrote clioutmsg xml to %s.',
                              conf_ismview_path)
        return out

    def ldbind(self, name, pool, ldn, size):
        """Bind an LD and attach a nickname to it."""
        errnum = ""
        cmd = ('iSMcfg ldbind -poolnumber %(poolnumber)d -ldn %(ldn)d '
               '-capacity %(capacity)d -immediate'
               % {'poolnumber': pool, 'ldn': ldn,
                  'capacity': size})
        out, err, status = self._execute(cmd, [0], False)
        errnum = err
        if status != 0:
            return False, errnum

        cmd = ('iSMcfg nickname -ldn %(ldn)d -newname %(newname)s '
               '-immediate'
               % {'ldn': ldn, 'newname': name})
        self._execute(cmd)
        return True, errnum

    def unbind(self, name):
        """Unbind an LD."""
        cmd = 'iSMcfg ldunbind -ldname %s' % name
        self._execute(cmd)

    def expand(self, ldn, capacity):
        """Expand a LD."""
        cmd = ('iSMcfg ldexpand -ldn %(ldn)d -capacity %(capacity)d '
               '-unit gb'
               % {'ldn': ldn, 'capacity': capacity})
        self._execute(cmd)

    def addldset_fc(self, ldsetname, connector):
        """Create new FC LD Set."""
        cmd = 'iSMcfg addldset -ldset LX:%s -type fc' % ldsetname
        out, err, status = self._execute(cmd, [0], False)
        if status != 0:
            return False
        for wwpn in connector['wwpns']:
            length = len(wwpn)
            setwwpn = '-'.join([wwpn[i:i + 4]
                                for i in range(0, length, 4)])
            setwwpn = setwwpn.upper()
            cmd = ('iSMcfg addldsetpath -ldset LX:%(name)s -path %(path)s'
                   % {'name': ldsetname, 'path': setwwpn})
            out, err, status = self._execute(cmd, [0], False)
            if status != 0:
                return False

        return True

    def addldset_iscsi(self, ldsetname, connector):
        """Create new iSCSI LD Set."""
        cmd = ('iSMcfg addldset -ldset LX:%s -type iscsi' % ldsetname)
        out, err, status = self._execute(cmd, [0], False)
        if status != 0:
            return False
        cmd = ('iSMcfg addldsetinitiator'
               ' -ldset LX:%(name)s -initiatorname %(initiator)s'
               % {'name': ldsetname, 'initiator': connector['initiator']})
        out, err, status = self._execute(cmd, [0], False)
        if status != 0:
            return False

        return True

    def addldsetld(self, ldset, ldname, lun=None):
        """Add an LD to specified LD Set."""
        if lun is None:
            cmd = ('iSMcfg addldsetld -ldset %(ldset)s '
                   '-ldname %(ldname)s'
                   % {'ldset': ldset, 'ldname': ldname})
            self._execute(cmd)
        else:
            cmd = ('iSMcfg addldsetld -ldset %(ldset)s -ldname %(ldname)s '
                   '-lun %(lun)d'
                   % {'ldset': ldset, 'ldname': ldname,
                      'lun': lun})
            self._execute(cmd)

    def delldsetld(self, ldset, ldname):
        """Delete an LD from specified LD Set."""
        rtn = True
        errnum = ""
        cmd = ('iSMcfg delldsetld -ldset %(ldset)s '
               '-ldname %(ldname)s'
               % {'ldset': ldset,
                  'ldname': ldname})
        out, err, status = self._execute(cmd, [0], False)
        errnum = err
        if status != 0:
            rtn = False
        return rtn, errnum

    def changeldname(self, ldn, new_name, old_name=None):
        """Rename nickname of LD."""
        if old_name is None:
            cmd = ('iSMcfg nickname -ldn %(ldn)d -newname %(newname)s '
                   '-immediate'
                   % {'ldn': ldn, 'newname': new_name})
            self._execute(cmd)
        else:
            cmd = ('iSMcfg nickname -ldname %(ldname)s '
                   '-newname %(newname)s'
                   % {'ldname': old_name,
                      'newname': new_name})
            self._execute(cmd)

    def setpair(self, mvname, rvname):
        """Set pair."""
        cmd = ('iSMrc_pair -pair -mv %(mv)s -mvflg ld '
               '-rv %(rv)s -rvflg ld'
               % {'mv': mvname, 'rv': rvname})
        self._execute(cmd)

        LOG.debug('Pair command completed. MV = %(mv)s RV = %(rv)s.',
                  {'mv': mvname, 'rv': rvname})

    def unpair(self, mvname, rvname, flag):
        """Unset pair."""
        if flag == 'normal':
            cmd = ('iSMrc_pair -unpair -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        elif flag == 'force':
            cmd = ('iSMrc_pair -unpair -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld -force all'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        else:
            LOG.debug('unpair flag ERROR. flag = %s', flag)

        LOG.debug('Unpair command completed. MV = %(mv)s, RV = %(rv)s.',
                  {'mv': mvname, 'rv': rvname})

    def replicate(self, mvname, rvname, flag):
        if flag == 'full':
            cmd = ('iSMrc_replicate -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld -nowait -cprange full '
                   '-cpmode bg'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        else:
            cmd = ('iSMrc_replicate -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld -nowait -cpmode bg'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)

        LOG.debug('Replicate command completed. MV = %(mv)s RV = %(rv)s.',
                  {'mv': mvname, 'rv': rvname})

    def separate(self, mvname, rvname, flag):
        """Separate for backup."""
        if flag == 'backup':
            cmd = ('iSMrc_separate -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld '
                   '-rvacc ro -rvuse complete -nowait'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        elif flag == 'restore' or flag == 'clone':
            cmd = ('iSMrc_separate -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld '
                   '-rvacc rw -rvuse immediate -nowait'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        elif flag == 'esv_restore' or flag == 'migrate':
            cmd = ('iSMrc_separate -mv %(mv)s -mvflg ld '
                   '-rv %(rv)s -rvflg ld '
                   '-rvacc rw -rvuse complete -nowait'
                   % {'mv': mvname, 'rv': rvname})
            self._execute(cmd)
        else:
            LOG.debug('separate flag ERROR. flag = %s', flag)

        LOG.debug('Separate command completed. MV = %(mv)s RV = %(rv)s.',
                  {'mv': mvname, 'rv': rvname})

    def query_MV_RV_status(self, ldname, rpltype):
        if rpltype == 'MV':
            cmd = ('iSMrc_query -mv %s -mvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "Sync State" ]]; '
                   'then builtin echo ${line:10};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
        elif rpltype == 'RV':
            cmd = ('iSMrc_query -rv %s -rvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "Sync State" ]]; '
                   'then builtin echo ${line:10};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
        else:
            LOG.debug('rpltype flag ERROR. rpltype = %s', rpltype)

        query_status = out.strip()
        return query_status

    def query_MV_RV_name(self, ldname, rpltype):
        if rpltype == 'MV':
            cmd = ('iSMrc_query -mv %s -mvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "LD Name" ]]; '
                   'then builtin echo ${line:7};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
            out = out.replace(ldname, "")
        elif rpltype == 'RV':
            cmd = ('iSMrc_query -rv %s -rvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "LD Name" ]]; '
                   'then builtin echo ${line:7};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
            out = out.replace(ldname, "")
        else:
            LOG.debug('rpltype flag ERROR. rpltype = %s', rpltype)

        query_name = out.strip()
        return query_name

    def query_MV_RV_diff(self, ldname, rpltype):
        if rpltype == 'MV':
            cmd = ('iSMrc_query -mv %s -mvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "Separate Diff" ]]; '
                   'then builtin echo ${line:13};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
        elif rpltype == 'RV':
            cmd = ('iSMrc_query -rv %s -rvflg ld | '
                   'while builtin read line;'
                   'do if [[ "$line" =~ "Separate Diff" ]]; '
                   'then builtin echo ${line:13};fi;'
                   'done' % ldname)
            out, err, status = self._execute(cmd)
        else:
            LOG.debug('rpltype flag ERROR. rpltype = %s', rpltype)

        query_status = out.strip()
        return query_status

    def backup_restore(self, volume_properties, unpairWait, canPairing=True):

        # Setting Pair.
        flag = 'full'
        if canPairing is True:
            self.setpair(volume_properties['mvname'][3:],
                         volume_properties['rvname'][3:])
        else:
            rv_diff = self.query_MV_RV_diff(volume_properties['rvname'][3:],
                                            'RV')
            rv_diff = int(rv_diff.replace('KB', ''), 10) // units.Ki
            if rv_diff != volume_properties['capacity']:
                flag = None

        # Replicate.
        self.replicate(volume_properties['mvname'][3:],
                       volume_properties['rvname'][3:], flag)

        # Separate.
        self.separate(volume_properties['mvname'][3:],
                      volume_properties['rvname'][3:],
                      volume_properties['flag'])

        unpairProc = unpairWait(volume_properties, self)
        unpairProc.run()

    def check_ld_existed_rplstatus(self, lds, ldname, snapshot, flag):

        if ldname not in lds:
            if flag == 'backup':
                LOG.debug('Volume Id not found. '
                          'LD name = %(name)s volume_id = %(id)s.',
                          {'name': ldname, 'id': snapshot.volume_id})
                raise exception.NotFound(_('Logical Disk does not exist.'))
            elif flag == 'restore':
                LOG.debug('Snapshot Id not found. '
                          'LD name = %(name)s snapshot_id = %(id)s.',
                          {'name': ldname, 'id': snapshot.id})
                raise exception.NotFound(_('Logical Disk does not exist.'))
            elif flag == 'delete':
                LOG.debug('LD `%(name)s` already unbound? '
                          'snapshot_id = %(id)s.',
                          {'name': ldname, 'id': snapshot.id})
                return None
            else:
                LOG.debug('check_ld_existed_rplstatus flag error flag = %s.',
                          flag)
                raise exception.NotFound(_('Logical Disk does not exist.'))

        ld = lds[ldname]

        if ld['RPL Attribute'] == 'IV':
            pass
        elif ld['RPL Attribute'] == 'MV':
            query_status = self.query_MV_RV_status(ldname[3:], 'MV')
            LOG.debug('query_status : %s.', query_status)
            if(query_status == 'separated'):
                # unpair.
                rvname = self.query_MV_RV_name(ldname[3:], 'MV')
                self.unpair(ldname[3:], rvname, 'force')
            else:
                msg = _('Specified Logical Disk %s has been copied.') % ldname
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        elif ld['RPL Attribute'] == 'RV':
            query_status = self.query_MV_RV_status(ldname[3:], 'RV')
            if query_status == 'separated':
                # unpair.
                mvname = self.query_MV_RV_name(ldname[3:], 'RV')
                self.unpair(mvname, ldname[3:], 'force')
            else:
                msg = _('Specified Logical Disk %s has been copied.') % ldname
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        return ld

    def get_pair_lds(self, ldname, lds):
        query_status = self.query_MV_RV_name(ldname[3:], 'MV')
        query_status = query_status.split('\n')
        query_status = [query for query in query_status if query != '']
        LOG.debug('query_status=%s.', query_status)

        pair_lds = {}
        for rvname in query_status:
            rvname = self._properties['ld_backupname_format'] % rvname
            if rvname not in lds:
                LOG.debug('LD `%s` is RDR pair?', rvname)
            else:
                ld = lds[rvname]
                ldn = ld['ldn']
                pair_lds[ldn] = ld

        LOG.debug('pair_lds=%s.', pair_lds)
        return pair_lds

    def snapshot_create(self, bvname, svname, poolnumber):
        """Snapshot create."""
        cmd = ('iSMcfg generationadd -bvname %(bvname)s '
               '-poolnumber %(poolnumber)d -count 1 '
               '-svname %(svname)s'
               % {'bvname': bvname,
                  'poolnumber': poolnumber,
                  'svname': svname})
        self._execute(cmd)

        cmd = ('iSMsc_create -bv %(bv)s -bvflg ld -sv %(sv)s '
               '-svflg ld'
               % {'bv': bvname[3:], 'sv': svname})
        self._execute(cmd)

    def snapshot_delete(self, bvname, svname):
        """Snapshot delete."""
        query_status = self.query_BV_SV_status(bvname[3:], svname)
        if query_status == 'snap/active':
            cmd = ('iSMsc_delete -bv %(bv)s -bvflg ld -sv %(sv)s '
                   '-svflg ld'
                   % {'bv': bvname[3:], 'sv': svname})
            self._execute(cmd)

            while True:
                query_status = self.query_BV_SV_status(bvname[3:], svname)
                if query_status == 'snap/deleting':
                    LOG.debug('Sleep 1 seconds Start')
                    time.sleep(1)
                else:
                    break
        else:
            LOG.debug('The snapshot data does not exist,'
                      ' because already forced deletion.'
                      ' bvname=%(bvname)s, svname=%(svname)s',
                      {'bvname': bvname, 'svname': svname})

        cmd = 'iSMcfg generationdel -bvname %s -count 1' % bvname
        self._execute(cmd)

    def snapshot_restore(self, bvname, svname):
        """Snapshot restore."""
        query_status = self.query_BV_SV_status(bvname[3:], svname[3:])
        if query_status == 'snap/active':
            cmd = ('iSMsc_restore -bv %(bv)s -bvflg ld -sv %(sv)s '
                   '-svflg ld -derivsv keep -nowait'
                   % {'bv': bvname[3:], 'sv': svname[3:]})
            self._execute(cmd)

            retry_count = 0
            while True:
                query_status = self.query_BV_SV_status(bvname[3:], svname[3:])
                if query_status == 'rst/exec':
                    # Restoration is in progress.
                    sleep_time = get_sleep_time_for_clone(retry_count)
                    LOG.debug('Sleep %d seconds Start', sleep_time)
                    time.sleep(sleep_time)
                    retry_count += 1
                elif query_status == 'snap/active':
                    # Restoration was successful.
                    break
                else:
                    # Restoration failed.
                    msg = (_('Failed to restore from snapshot. '
                             'bvname=%(bvname)s, svname=%(svname)s, '
                             'status=%(status)s') %
                           {'bvname': bvname, 'svname': svname,
                            'status': query_status})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
        else:
            msg = (_('The snapshot does not exist or is '
                     'not in snap/active status. '
                     'bvname=%(bvname)s, svname=%(svname)s, '
                     'status=%(status)s') %
                   {'bvname': bvname, 'svname': svname,
                    'status': query_status})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def query_BV_SV_status(self, bvname, svname):
        cmd = ('iSMsc_query -bv %(bv)s -bvflg ld -sv %(sv)s -svflg ld '
               '-summary | '
               'while builtin read line;do '
               'if [[ "$line" =~ "%(line)s" ]]; '
               'then builtin echo "$line";fi;done'
               % {'bv': bvname, 'sv': svname, 'line': svname})
        out, err, status = self._execute(cmd)

        delimiter = ') '
        start = out.find(delimiter)
        if start == -1:
            return None
        start += len(delimiter)
        query_status = out[start:].split(' ')[0]
        LOG.debug('snap/state:%s.', query_status)
        return query_status

    def get_bvname(self, svname):
        cmd = ('iSMsc_query -sv %s -svflg ld -summary | '
               'while builtin read line;do '
               'if [[ "$line" =~ "LD Name" ]]; '
               'then builtin echo "$line";fi;done'
               % svname[3:])
        out, err, status = self._execute(cmd)

        query_status = out[15:39].strip()
        return query_status

    def set_io_limit(self, ldname, qos_params, force_delete=True):
        upper = qos_params['upperlimit']
        lower = qos_params['lowerlimit']
        report = qos_params['upperreport']
        if upper is None and lower is None and report is None:
            return
        cmd = 'iSMioc setlimit -ldname %s' % ldname
        if upper is not None:
            cmd += ' -upperlimit %d' % upper
        if lower is not None:
            cmd += ' -lowerlimit %d' % lower
        if report is not None:
            cmd += ' -upperreport %s' % report
        try:
            self._execute(cmd)
        except Exception:
            with excutils.save_and_reraise_exception():
                if force_delete:
                    self.unbind(ldname)

    def lvbind(self, bvname, lvname, lvnumber):
        """Link Volume create."""
        cmd = ('iSMcfg lvbind -bvname %(bvname)s '
               '-lvn %(lvnumber)d -lvname %(lvname)s'
               % {'bvname': bvname,
                  'lvnumber': lvnumber,
                  'lvname': lvname})
        self._execute(cmd)

    def lvunbind(self, lvname):
        """Link Volume delete."""
        cmd = ('iSMcfg lvunbind -ldname %(lvname)s'
               % {'lvname': lvname})
        self._execute(cmd)

    def lvlink(self, svname, lvname):
        """Link to snapshot volume."""
        cmd = ('iSMsc_link -lv %(lvname)s -lvflg ld '
               '-sv %(svname)s -svflg ld -lvacc ro'
               % {'lvname': lvname,
                  'svname': svname})
        self._execute(cmd)

    def lvunlink(self, lvname):
        """Unlink from snapshot volume."""
        cmd = ('iSMsc_unlink -lv %(lvname)s -lvflg ld'
               % {'lvname': lvname})
        self._execute(cmd)

    def cvbind(self, poolnumber, cvnumber):
        """Create Control Volume."""
        cmd = ('iSMcfg ldbind -poolnumber %(poolnumber)d '
               '-ldattr cv -ldn %(cvnumber)d'
               % {'poolnumber': poolnumber,
                  'cvnumber': cvnumber})
        self._execute(cmd)


class UnpairWait(object):

    def __init__(self, volume_properties, cli):
        super(UnpairWait, self).__init__()
        self._volume_properties = volume_properties
        self._mvname = volume_properties['mvname'][3:]
        self._rvname = volume_properties['rvname'][3:]
        self._mvID = volume_properties['mvid']
        self._rvID = volume_properties['rvid']
        self._flag = volume_properties['flag']
        self._context = volume_properties['context']
        self._cli = cli
        self._local_conf = self._cli._properties

    def _wait(self, unpair=True):
        timeout = self._local_conf['thread_timeout'] * 24
        start_time = time.time()
        retry_count = 0
        while True:
            cur_time = time.time()
            if (cur_time - start_time) > timeout:
                raise exception.APITimeout(_('UnpairWait wait timeout.'))

            sleep_time = get_sleep_time_for_clone(retry_count)
            LOG.debug('Sleep %d seconds Start', sleep_time)
            time.sleep(sleep_time)
            retry_count += 1

            query_status = self._cli.query_MV_RV_status(self._rvname, 'RV')
            if query_status == 'separated':
                if unpair is True:
                    self._cli.unpair(self._mvname, self._rvname, 'normal')
                break
            elif query_status == 'sep/exec':
                continue
            else:
                LOG.debug('iSMrc_query command result abnormal.'
                          'Query status = %(status)s, RV = %(rv)s.',
                          {'status': query_status, 'rv': self._rvname})
                break

    def run(self):
        try:
            self._execute()
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.debug('UnpairWait Unexpected error. '
                          'exception=%(exception)s, MV = %(mv)s, RV = %(rv)s.',
                          {'exception': traceback.format_exc(),
                           'mv': self._mvname, 'rv': self._rvname})

    def _execute(self):
        pass


class UnpairWaitForRestore(UnpairWait):
    def __init__(self, volume_properties, cli):
        super(UnpairWaitForRestore, self).__init__(volume_properties, cli)

        self._rvldn = None
        if ('rvldn' in volume_properties and
                volume_properties['rvldn'] is not None):
            self._rvldn = volume_properties['rvldn']

        self._rvcapacity = None
        if ('rvcapacity' in volume_properties and
                volume_properties['rvcapacity'] is not None):
            self._rvcapacity = volume_properties['rvcapacity']

    def _execute(self):
        LOG.debug('UnpairWaitForRestore start.')

        self._wait(True)

        if self._rvcapacity is not None:
            try:
                self._cli.expand(self._rvldn, self._rvcapacity)
            except exception.CinderException:
                with excutils.save_and_reraise_exception():
                    LOG.debug('UnpairWaitForDDRRestore expand error. '
                              'exception=%(exception)s, '
                              'MV = %(mv)s, RV = %(rv)s.',
                              {'exception': traceback.format_exc(),
                               'mv': self._mvname, 'rv': self._rvname})


class UnpairWaitForClone(UnpairWait):
    def __init__(self, volume_properties, cli):
        super(UnpairWaitForClone, self).__init__(volume_properties, cli)

        self._rvldn = None
        if ('rvldn' in volume_properties and
                volume_properties['rvldn'] is not None):
            self._rvldn = volume_properties['rvldn']

        self._rvcapacity = None
        if ('rvcapacity' in volume_properties and
                volume_properties['rvcapacity'] is not None):
            self._rvcapacity = volume_properties['rvcapacity']

    def _execute(self):
        LOG.debug('UnpairWaitForClone start.')

        self._wait(True)

        if self._rvcapacity is not None:
            try:
                self._cli.expand(self._rvldn, self._rvcapacity)
            except exception.CinderException:
                with excutils.save_and_reraise_exception():
                    LOG.debug('UnpairWaitForClone expand error. '
                              'exception=%(exception)s, '
                              'MV = %(mv)s, RV = %(rv)s.',
                              {'exception': traceback.format_exc(),
                               'mv': self._mvname, 'rv': self._rvname})


class UnpairWaitForMigrate(UnpairWait):
    def __init__(self, volume_properties, cli):
        super(UnpairWaitForMigrate, self).__init__(volume_properties, cli)

    def _execute(self):
        LOG.debug('UnpairWaitForMigrate start.')

        self._wait(True)

        self._cli.unbind(self._volume_properties['mvname'])
        self._cli.changeldname(None, self._volume_properties['mvname'],
                               self._volume_properties['rvname'])


class UnpairWaitForDDRRestore(UnpairWaitForRestore):
    def __init__(self, volume_properties, cli):
        super(UnpairWaitForDDRRestore, self).__init__(volume_properties, cli)

        self._prev_mvname = None
        if ('prev_mvname' in volume_properties and
                volume_properties['prev_mvname'] is not None):
            self._prev_mvname = volume_properties['prev_mvname'][3:]

    def _execute(self):
        LOG.debug('UnpairWaitForDDRRestore start.')

        self._wait(True)

        if self._rvcapacity is not None:
            try:
                self._cli.expand(self._rvldn, self._rvcapacity)
            except exception.CinderException:
                with excutils.save_and_reraise_exception():
                    LOG.debug('UnpairWaitForDDRRestore expand error. '
                              'exception=%(exception)s, '
                              'MV = %(mv)s, RV = %(rv)s.',
                              {'exception': traceback.format_exc(),
                               'mv': self._mvname, 'rv': self._rvname})

        if self._prev_mvname is not None:
            self._cli.setpair(self._prev_mvname, self._mvname)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Fakes For Scheduler tests.
"""
import copy

from oslo_utils import timeutils
from oslo_utils import uuidutils

from cinder.scheduler import filter_scheduler
from cinder.scheduler import host_manager
from cinder.volume import volume_utils


UTC_NOW = timeutils.utcnow()

SERVICE_STATES = {
    'host1': {'total_capacity_gb': 1024,
              'free_capacity_gb': 1024,
              'allocated_capacity_gb': 0,
              'provisioned_capacity_gb': 0,
              'max_over_subscription_ratio': '1.0',
              'thin_provisioning_support': False,
              'thick_provisioning_support': True,
              'reserved_percentage': 10,
              'volume_backend_name': 'lvm1',
              'timestamp': UTC_NOW,
              'multiattach': True,
              'online_extend_support': True,
              'uuid': 'a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'},
    'host2': {'total_capacity_gb': 2048,
              'free_capacity_gb': 300,
              'allocated_capacity_gb': 1748,
              'provisioned_capacity_gb': 1748,
              'max_over_subscription_ratio': '1.5',
              'thin_provisioning_support': True,
              'thick_provisioning_support': False,
              'reserved_percentage': 10,
              'volume_backend_name': 'lvm2',
              'timestamp': UTC_NOW,
              'online_extend_support': False,
              'uuid': '4200b32b-0bf9-436c-86b2-0675f6ac218e'},
    'host3': {'total_capacity_gb': 512,
              'free_capacity_gb': 256,
              'allocated_capacity_gb': 256,
              'provisioned_capacity_gb': 256,
              'max_over_subscription_ratio': '2.0',
              'thin_provisioning_support': False,
              'thick_provisioning_support': True,
              'reserved_percentage': 0,
              'volume_backend_name': 'lvm3',
              'timestamp': UTC_NOW,
              'uuid': '6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7'},
    'host4': {'total_capacity_gb': 2048,
              'free_capacity_gb': 200,
              'allocated_capacity_gb': 1848,
              'provisioned_capacity_gb': 2047,
              'max_over_subscription_ratio': '1.0',
              'thin_provisioning_support': True,
              'thick_provisioning_support': False,
              'reserved_percentage': 5,
              'volume_backend_name': 'lvm4',
              'timestamp': UTC_NOW,
              'consistent_group_snapshot_enabled': True,
              'uuid': '18417850-2ca9-43d1-9619-ae16bfb0f655'},
    'host5': {'total_capacity_gb': 'infinite',
              'free_capacity_gb': 'unknown',
              'allocated_capacity_gb': 1548,
              'provisioned_capacity_gb': 1548,
              'max_over_subscription_ratio': '1.0',
              'thin_provisioning_support': True,
              'thick_provisioning_support': False,
              'reserved_percentage': 5,
              'timestamp': UTC_NOW,
              'uuid': 'f838f35c-4035-464f-9792-ce60e390c13d'},
}

SERVICE_STATES_WITH_POOLS = {
    'host1@BackendA': {
        'uuid': 'a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824',
        'replication_enabled': False,
        'driver_version': '1.0.0',
        'volume_backend_name': 'BackendA',
        'pools': [
            {
                'total_capacity_gb': 1024,
                'free_capacity_gb': 1024,
                'allocated_capacity_gb': 0,
                'provisioned_capacity_gb': 0,
                'max_over_subscription_ratio': '1.0',
                'thin_provisioning_support': False,
                'thick_provisioning_support': True,
                'reserved_percentage': 15,
                'pool_name': 'openstack_iscsi_1',
            },
            {
                'total_capacity_gb': 2048,
                'free_capacity_gb': 1008,
                'allocated_capacity_gb': 0,
                'provisioned_capacity_gb': 0,
                'max_over_subscription_ratio': '1.0',
                'thin_provisioning_support': True,
                'thick_provisioning_support': False,
                'reserved_percentage': 15,
                'pool_name': 'openstack_iscsi_2',
            },

        ],
        'storage_protocol': 'iSCSI',
        'timestamp': UTC_NOW,
    },
    'host1@BackendB': {
        'replication_enabled': True,
        'driver_version': '1.5.0',
        'volume_backend_name': 'BackendB',
        'uuid': '4200b32b-0bf9-436c-86b2-0675f6ac218e',
        'pools': [
            {
                'total_capacity_gb': 2048,
                'free_capacity_gb': 300,
                'allocated_capacity_gb': 1748,
                'provisioned_capacity_gb': 1748,
                'max_over_subscription_ratio': '1.5',
                'thin_provisioning_support': True,
                'thick_provisioning_support': False,
                'reserved_percentage': 10,
                'pool_name': 'openstack_nfs_1',
            },
            {
                'total_capacity_gb': 512,
                'free_capacity_gb': 256,
                'allocated_capacity_gb': 256,
                'provisioned_capacity_gb': 256,
                'max_over_subscription_ratio': '2.0',
                'thin_provisioning_support': True,
                'thick_provisioning_support': False,
                'reserved_percentage': 10,
                'pool_name': 'openstack_nfs_2',
            },

        ],
        'storage_protocol': 'nfs',
        'timestamp': UTC_NOW,
    },
    'host2@BackendX': {
        'replication_enabled': False,
        'driver_version': '3.5.1',
        'total_capacity_gb': 512,
        'free_capacity_gb': 256,
        'allocated_capacity_gb': 256,
        'provisioned_capacity_gb': 256,
        'max_over_subscription_ratio': '2.0',
        'thin_provisioning_support': False,
        'thick_provisioning_support': True,
        'reserved_percentage': 0,
        'volume_backend_name': 'BackendX',
        'storage_protocol': 'iSCSI',
        'timestamp': UTC_NOW,
        'uuid': '6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7'
    },
    'host3@BackendY': {
        'replication_enabled': True,
        'driver_version': '1.5.0',
        'volume_backend_name': 'BackendY',
        'uuid': '18417850-2ca9-43d1-9619-ae16bfb0f655',
        'pools': [
            {
                'total_capacity_gb': 'infinite',
                'free_capacity_gb': 'unknown',
                'allocated_capacity_gb': 170,
                'provisioned_capacity_gb': 170,
                'max_over_subscription_ratio': '1.0',
                'thin_provisioning_support': False,
                'thick_provisioning_support': True,
                'QoS_support': True,
                'reserved_percentage': 0,
                'pool_name': 'openstack_fcp_1',
            },
            {
                'total_capacity_gb': 'infinite',
                'free_capacity_gb': 'unknown',
                'allocated_capacity_gb': 1548,
                'provisioned_capacity_gb': 1548,
                'max_over_subscription_ratio': '1.0',
                'thin_provisioning_support': True,
                'thick_provisioning_support': False,
                'QoS_support': True,
                'reserved_percentage': 0,
                'pool_name': 'openstack_fcp_2',
            },

        ],
        'storage_protocol': 'fc',
        'timestamp': UTC_NOW,
    }
}


class FakeFilterScheduler(filter_scheduler.FilterScheduler):
    def __init__(self, *args, **kwargs):
        super(FakeFilterScheduler, self).__init__(*args, **kwargs)
        self.host_manager = host_manager.HostManager()


class FakeHostManager(host_manager.HostManager):
    def __init__(self, multibackend_with_pools=False):
        super(FakeHostManager, self).__init__()

        self.service_states = copy.deepcopy(
            SERVICE_STATES_WITH_POOLS if multibackend_with_pools
            else SERVICE_STATES
        )


class FakeBackendState(host_manager.BackendState):
    def __init__(self, host, attribute_dict):
        super(FakeBackendState, self).__init__(host, None)
        for (key, val) in attribute_dict.items():
            setattr(self, key, val)


class FakeNovaClient(object):
    class Server(object):
        def __init__(self, host):
            self.uuid = uuidutils.generate_uuid()
            self.host = host
            setattr(self, 'OS-EXT-SRV-ATTR:host', host)

    class ServerManager(object):
        def __init__(self):
            self._servers = []

        def create(self, host):
            self._servers.append(FakeNovaClient.Server(host))
            return self._servers[-1].uuid

        def get(self, server_uuid):
            for s in self._servers:
                if s.uuid == server_uuid:
                    return s
            return None

        def list(self, detailed=True, search_opts=None):
            matching = list(self._servers)
            if search_opts:
                for opt, val in search_opts.items():
                    matching = [m for m in matching
                                if getattr(m, opt, None) == val]
            return matching

    class ListExtResource(object):
        def __init__(self, ext_name):
            self.name = ext_name

    class ListExtManager(object):
        def __init__(self, ext_srv_attr=True):
            self.ext_srv_attr = ext_srv_attr

        def show_all(self):
            if self.ext_srv_attr:
                return [
                    FakeNovaClient.ListExtResource('ExtendedServerAttributes')]
            return []

    def __init__(self, ext_srv_attr=True):
        self.servers = FakeNovaClient.ServerManager()
        self.list_extensions = FakeNovaClient.ListExtManager(
            ext_srv_attr=ext_srv_attr)


def mock_host_manager_db_calls(mock_obj, backends_with_pools=False,
                               disabled=None):
    service_states = (
        SERVICE_STATES_WITH_POOLS if backends_with_pools else SERVICE_STATES
    )
    services = []
    az_map = {
        'host1': 'zone1',
        'host2': 'zone1',
        'host3': 'zone2',
        'host4': 'zone3',
        'host5': 'zone3',
    }
    sid = 0
    for svc, state in service_states.items():
        sid += 1
        services.append(
            {
                'id': sid,
                'host': svc,
                'availability_zone': az_map[volume_utils.extract_host(svc,
                                                                      'host')],
                'topic': 'volume',
                'disabled': False,
                'updated_at': timeutils.utcnow(),
                'uuid': state.get('uuid', uuidutils.generate_uuid()),
            }
        )

    if disabled is None:
        mock_obj.return_value = services
    else:
        mock_obj.return_value = [service for service in services
                                 if service['disabled'] == disabled]



#    (c) Copyright 2016 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Unit tests for brcd fc zone client cli."""
from unittest import mock

from oslo_concurrency import processutils

from cinder import exception
from cinder.tests.unit import test
from cinder.zonemanager.drivers.brocade import (brcd_fc_zone_client_cli
                                                as client_cli)
from cinder.zonemanager.drivers.brocade import exception as b_exception
import cinder.zonemanager.drivers.brocade.fc_zone_constants as zone_constant


nsshow = '20:1a:00:05:1e:e8:e3:29'
switch_data = [' N 011a00;2,3;20:1a:00:05:1e:e8:e3:29;\
                 20:1a:00:05:1e:e8:e3:29;na',
               '    Fabric Port Name: 20:1a:00:05:1e:e8:e3:29']
cfgactvshow = ['Effective configuration:\n',
               ' cfg:\tOpenStack_Cfg\t\n',
               ' zone:\topenstack50060b0000c26604201900051ee8e329\t\n',
               '\t\t50:06:0b:00:00:c2:66:04\n',
               '\t\t20:19:00:05:1e:e8:e3:29\n']
active_zoneset = {
    'zones': {
        'openstack50060b0000c26604201900051ee8e329':
        ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']},
    'active_zone_config': 'OpenStack_Cfg'}
active_zoneset_multiple_zones = {
    'zones': {
        'openstack50060b0000c26604201900051ee8e329':
        ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29'],
        'openstack50060b0000c26602201900051ee8e327':
        ['50:06:0b:00:00:c2:66:02', '20:19:00:05:1e:e8:e3:27']},
    'active_zone_config': 'OpenStack_Cfg'}
new_zone_memb_same = {
    'openstack50060b0000c26604201900051ee8e329':
    ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']}
new_zone_memb_not_same = {
    'openstack50060b0000c26604201900051ee8e330':
    ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:30']}
new_zone = {'openstack10000012345678902001009876543210':
            ['10:00:00:12:34:56:78:90', '20:01:00:98:76:54:32:10']}
new_zones = {'openstack10000012345678902001009876543210':
             ['10:00:00:12:34:56:78:90', '20:01:00:98:76:54:32:10'],
             'openstack10000011111111112001001111111111':
             ['10:00:00:11:11:11:11:11', '20:01:00:11:11:11:11:11']}
zone_names_to_delete = 'openstack50060b0000c26604201900051ee8e329'
supported_firmware = ['Kernel: 2.6', 'Fabric OS:  v7.0.1']
unsupported_firmware = ['Fabric OS:  v6.2.1']


class TestBrcdFCZoneClientCLI(client_cli.BrcdFCZoneClientCLI, test.TestCase):

    # override some of the functions
    def __init__(self, *args, **kwargs):
        test.TestCase.__init__(self, *args, **kwargs)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_get_switch_info')
    def test_get_active_zone_set(self, get_switch_info_mock):
        cmd_list = [zone_constant.GET_ACTIVE_ZONE_CFG]
        get_switch_info_mock.return_value = cfgactvshow
        active_zoneset_returned = self.get_active_zone_set()
        get_switch_info_mock.assert_called_once_with(cmd_list)
        self.assertDictEqual(active_zoneset, active_zoneset_returned)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test_get_active_zone_set_ssh_error(self, run_ssh_mock):
        run_ssh_mock.side_effect = processutils.ProcessExecutionError
        self.assertRaises(b_exception.BrocadeZoningCliException,
                          self.get_active_zone_set)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'get_active_zone_set')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_cfg_save')
    def test_add_zones_new_zone_no_activate(self, cfg_save_mock,
                                            apply_zone_change_mock,
                                            get_active_zs_mock):
        get_active_zs_mock.return_value = active_zoneset
        self.add_zones(new_zones, False, None)
        self.assertEqual(1, get_active_zs_mock.call_count)
        self.assertEqual(3, apply_zone_change_mock.call_count)
        cfg_save_mock.assert_called_once_with()

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'get_active_zone_set')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'activate_zoneset')
    def test_add_zones_new_zone_activate(self, activate_zoneset_mock,
                                         apply_zone_change_mock,
                                         get_active_zs_mock):
        get_active_zs_mock.return_value = active_zoneset
        self.add_zones(new_zone, True, active_zoneset)
        self.assertEqual(2, apply_zone_change_mock.call_count)
        activate_zoneset_mock.assert_called_once_with(
            active_zoneset['active_zone_config'])

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'get_active_zone_set')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'activate_zoneset')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    def test_update_zone_exists_memb_same(self, apply_zone_change_mock,
                                          activate_zoneset_mock,
                                          get_active_zs_mock):
        get_active_zs_mock.return_value = active_zoneset
        self.update_zones(new_zone_memb_same, True, zone_constant.ZONE_ADD,
                          active_zoneset)
        self.assertEqual(1, apply_zone_change_mock.call_count)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'get_active_zone_set')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'activate_zoneset')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    def test_update_zone_exists_memb_not_same(self, apply_zone_change_mock,
                                              activate_zoneset_mock,
                                              get_active_zs_mock):
        get_active_zs_mock.return_value = active_zoneset
        self.update_zones(new_zone_memb_not_same, True,
                          zone_constant.ZONE_ADD, active_zoneset)
        self.assertEqual(1, apply_zone_change_mock.call_count)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'get_active_zone_set')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'activate_zoneset')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    def test_add_zone_all_exists_memb_not_same(self, apply_zone_change_mock,
                                               activate_zoneset_mock,
                                               get_active_zs_mock):

        self.add_zones(new_zone_memb_not_same, True, active_zoneset)
        call_args = apply_zone_change_mock.call_args[0][0]
        self.assertEqual(0, get_active_zs_mock.call_count)
        self.assertEqual(2, apply_zone_change_mock.call_count)
        self.assertIn(zone_constant.CFG_ADD.strip(), call_args)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_ssh_execute')
    def test_activate_zoneset(self, ssh_execute_mock):
        ssh_execute_mock.return_value = True
        return_value = self.activate_zoneset('zoneset1')
        self.assertTrue(return_value)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_ssh_execute')
    def test_deactivate_zoneset(self, ssh_execute_mock):
        ssh_execute_mock.return_value = True
        return_value = self.deactivate_zoneset()
        self.assertTrue(return_value)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_cfg_save')
    def test_delete_zones_activate_false(self, cfg_save_mock,
                                         apply_zone_change_mock):
        with mock.patch.object(self, '_zone_delete') as zone_delete_mock:
            self.delete_zones(zone_names_to_delete, False,
                              active_zoneset_multiple_zones)
            self.assertEqual(1, apply_zone_change_mock.call_count)
            zone_delete_mock.assert_called_once_with(zone_names_to_delete)
            cfg_save_mock.assert_called_once_with()

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'activate_zoneset')
    def test_delete_zones_activate_true(self, activate_zs_mock,
                                        apply_zone_change_mock):
        with mock.patch.object(self, '_zone_delete') \
                as zone_delete_mock:
            self.delete_zones(zone_names_to_delete, True,
                              active_zoneset_multiple_zones)
            self.assertEqual(1, apply_zone_change_mock.call_count)
            zone_delete_mock.assert_called_once_with(zone_names_to_delete)
            activate_zs_mock.assert_called_once_with(
                active_zoneset['active_zone_config'])

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_get_switch_info')
    def test_get_nameserver_info(self, get_switch_info_mock):
        ns_info_list_expected = ['20:1a:00:05:1e:e8:e3:29']
        get_switch_info_mock.return_value = (switch_data)
        ns_info_list = self.get_nameserver_info()
        self.assertEqual(ns_info_list_expected, ns_info_list)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test_get_nameserver_info_ssh_error(self, run_ssh_mock):
        run_ssh_mock.side_effect = processutils.ProcessExecutionError
        self.assertRaises(b_exception.BrocadeZoningCliException,
                          self.get_nameserver_info)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_ssh_execute')
    def test__cfg_save(self, ssh_execute_mock):
        cmd_list = [zone_constant.CFG_SAVE]
        self._cfg_save()
        ssh_execute_mock.assert_called_once_with(cmd_list, True, 1)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    def test__zone_delete(self, apply_zone_change_mock):
        zone_name = 'testzone'
        cmd_list = ['zonedelete', '"testzone"']
        self._zone_delete(zone_name)
        apply_zone_change_mock.assert_called_once_with(cmd_list)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, 'apply_zone_change')
    def test__cfg_trans_abort(self, apply_zone_change_mock):
        cmd_list = [zone_constant.CFG_ZONE_TRANS_ABORT]
        with mock.patch.object(self, '_is_trans_abortable') \
                as is_trans_abortable_mock:
            is_trans_abortable_mock.return_value = True
            self._cfg_trans_abort()
            is_trans_abortable_mock.assert_called_once_with()
            apply_zone_change_mock.assert_called_once_with(cmd_list)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test__is_trans_abortable_true(self, run_ssh_mock):
        cmd_list = [zone_constant.CFG_SHOW_TRANS]
        run_ssh_mock.return_value = (Stream(zone_constant.TRANS_ABORTABLE),
                                     None)
        data = self._is_trans_abortable()
        self.assertTrue(data)
        run_ssh_mock.assert_called_once_with(cmd_list, True, 1)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test__is_trans_abortable_ssh_error(self, run_ssh_mock):
        run_ssh_mock.return_value = (Stream(), Stream())
        self.assertRaises(b_exception.BrocadeZoningCliException,
                          self._is_trans_abortable)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test__is_trans_abortable_false(self, run_ssh_mock):
        cmd_list = [zone_constant.CFG_SHOW_TRANS]
        cfgtransshow = 'There is no outstanding zoning transaction'
        run_ssh_mock.return_value = (Stream(cfgtransshow), None)
        data = self._is_trans_abortable()
        self.assertFalse(data)
        run_ssh_mock.assert_called_once_with(cmd_list, True, 1)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test_apply_zone_change(self, run_ssh_mock):
        cmd_list = [zone_constant.CFG_SAVE]
        run_ssh_mock.return_value = (None, None)
        self.apply_zone_change(cmd_list)
        run_ssh_mock.assert_called_once_with(cmd_list, True, 1)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_run_ssh')
    def test__get_switch_info(self, run_ssh_mock):
        cmd_list = [zone_constant.NS_SHOW]
        nsshow_list = [nsshow]
        run_ssh_mock.return_value = (Stream(nsshow), Stream())
        switch_data = self._get_switch_info(cmd_list)
        self.assertEqual(nsshow_list, switch_data)
        run_ssh_mock.assert_called_once_with(cmd_list, True, 1)

    def test__parse_ns_output(self):
        invalid_switch_data = [' N 011a00;20:1a:00:05:1e:e8:e3:29']
        expected_wwn_list = ['20:1a:00:05:1e:e8:e3:29']
        return_wwn_list = self._parse_ns_output(switch_data)
        self.assertEqual(expected_wwn_list, return_wwn_list)
        self.assertRaises(exception.InvalidParameterValue,
                          self._parse_ns_output, invalid_switch_data)

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_execute_shell_cmd')
    def test_is_supported_firmware(self, exec_shell_cmd_mock):
        exec_shell_cmd_mock.return_value = (supported_firmware, None)
        self.assertTrue(self.is_supported_firmware())

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_execute_shell_cmd')
    def test_is_supported_firmware_invalid(self, exec_shell_cmd_mock):
        exec_shell_cmd_mock.return_value = (unsupported_firmware, None)
        self.assertFalse(self.is_supported_firmware())

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_execute_shell_cmd')
    def test_is_supported_firmware_no_ssh_response(self, exec_shell_cmd_mock):
        exec_shell_cmd_mock.return_value = (None, Stream())
        self.assertFalse(self.is_supported_firmware())

    @mock.patch.object(client_cli.BrcdFCZoneClientCLI, '_execute_shell_cmd')
    def test_is_supported_firmware_ssh_error(self, exec_shell_cmd_mock):
        exec_shell_cmd_mock.side_effect = processutils.ProcessExecutionError
        self.assertRaises(b_exception.BrocadeZoningCliException,
                          self.is_supported_firmware)


class Channel(object):
    def recv_exit_status(self):
        return 0


class Stream(object):
    def __init__(self, buffer=''):
        self.buffer = buffer
        self.channel = Channel()

    def readlines(self):
        return self.buffer

    def splitlines(self):
        return self.buffer.splitlines()

    def close(self):
        pass

    def flush(self):
        self.buffer = ''



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

import requests

from cinder import context
from cinder.tests.unit import test
from cinder.volume import configuration
from cinder.volume.drivers.dell_emc.powerstore import driver
from cinder.volume.drivers.dell_emc.powerstore import options


class MockResponse(requests.Response):
    def __init__(self, content=None, rc=200):
        super(MockResponse, self).__init__()

        if content is None:
            content = []
        if isinstance(content, str):
            content = content.encode()
        self._content = content
        self.request = mock.MagicMock()
        self.status_code = rc

    def json(self, **kwargs):
        if isinstance(self._content, bytes):
            return super(MockResponse, self).json(**kwargs)
        return self._content

    @property
    def text(self):
        if not isinstance(self._content, bytes):
            return json.dumps(self._content)
        return super(MockResponse, self).text


class TestPowerStoreDriver(test.TestCase):
    def setUp(self):
        super(TestPowerStoreDriver, self).setUp()
        self.context = context.RequestContext('fake', 'fake', auth_token=True)
        self.configuration = configuration.Configuration(
            options.POWERSTORE_OPTS,
            configuration.SHARED_CONF_GROUP
        )
        self._set_overrides()
        self.driver = driver.PowerStoreDriver(configuration=self.configuration)
        self.driver.do_setup({})
        self.iscsi_driver = self.driver

        self._override_shared_conf("storage_protocol", override="FC")
        self.fc_driver = driver.PowerStoreDriver(
            configuration=self.configuration
        )
        self.fc_driver.do_setup({})

    def _override_shared_conf(self, *args, **kwargs):
        return self.override_config(*args,
                                    **kwargs,
                                    group=configuration.SHARED_CONF_GROUP)

    def _set_overrides(self):
        # Override the defaults to fake values
        self._override_shared_conf("san_ip", override="127.0.0.1")
        self._override_shared_conf("san_login", override="test")
        self._override_shared_conf("san_password", override="test")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.volume.targets import iscsi


class FakeTarget(iscsi.ISCSITarget):
    VERSION = '0.1'

    def _get_target_and_lun(self, context, volume):
        return(0, 0)

    def create_iscsi_target(self, name, tid, lun, path,
                            chap_auth, **kwargs):
        pass

    def remove_iscsi_target(self, tid, lun, vol_id, vol_name, **kwargs):
        pass

    def _get_iscsi_target(self, context, vol_id):
        pass

    def _get_target(self, iqn):
        pass



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
from unittest import mock

from oslo_utils import units as o_units

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume.drivers.open_e import iscsi
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom

UUID_1 = '12345678-1234-1234-1234-000000000001'
UUID_2 = '12345678-1234-1234-1234-000000000002'
UUID_3 = '12345678-1234-1234-1234-000000000003'
UUID_4 = '12345678-1234-1234-1234-000000000004'

CONFIG_OK = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '128K'
}

CONFIG_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '64K'
}

CONFIG_BAD_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '61K'
}

CONFIG_BACKEND_NAME = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

CONFIG_MULTI_HOST = {
    'san_hosts': ['192.168.0.2', '192.168.0.3'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

SNAPSHOTS_CASCADE_1 = [
    {"name": jcom.sname(UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_1)},
    {"name": jcom.sname(UUID_2),
     "clones": "Pool-0/" + jcom.sname(UUID_2)},
    {"name": jcom.sname(UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_3)}]

SNAPSHOTS_CASCADE_2 = [
    {"name": jcom.sname(UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_1)},
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_3)}]

SNAPSHOTS_CASCADE_3 = [
    {"name": jcom.vname(UUID_4),
     "clones": "Pool-0/" + jcom.vname(UUID_4)}]

SNAPSHOTS_EMPTY = []

SNAPSHOTS_CLONE = [
    {"name": jcom.vname(UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_1)}]

SNAPSHOTS_GARBAGE = [
    {"name": jcom.sname(UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_2),
     "clones": ""}]

SNAPSHOTS_RECURSIVE_1 = [
    {"name": jcom.sname(UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_1)},
    {"name": jcom.sname(UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]

SNAPSHOTS_RECURSIVE_CHAIN_1 = [
    {"name": jcom.sname(UUID_3),
     "clones": "Pool-0/" + jcom.hidden(UUID_3)}]

SNAPSHOTS_RECURSIVE_CHAIN_2 = [
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]


def get_jdss_exceptions():

    out = [jexc.JDSSException(reason="Testing"),
           jexc.JDSSRESTException(request="ra request", reason="Testing"),
           jexc.JDSSRESTProxyException(host="test_host", reason="Testing"),
           jexc.JDSSResourceNotFoundException(res="test_resource"),
           jexc.JDSSVolumeNotFoundException(volume="test_volume"),
           jexc.JDSSSnapshotNotFoundException(snapshot="test_snapshot"),
           jexc.JDSSResourceExistsException(res="test_resource"),
           jexc.JDSSSnapshotExistsException(snapshot="test_snapshot"),
           jexc.JDSSVolumeExistsException(volume="test_volume"),
           jexc.JDSSSnapshotIsBusyException(snapshot="test_snapshot")]

    return out


class TestOpenEJovianDSSDriver(test.TestCase):

    def get_driver(self, config):
        ctx = context.get_admin_context()

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None
        cfg.get = lambda val, default: config.get(val, default)

        jdssd = iscsi.JovianISCSIDriver()

        jdssd.configuration = cfg
        lib_to_patch = ('cinder.volume.drivers.open_e.jovian_common.rest.'
                        'JovianRESTAPI')
        with mock.patch(lib_to_patch) as ra:
            ra.is_pool_exists.return_value = True
            jdssd.do_setup(ctx)
        jdssd.ra = mock.Mock()
        return jdssd, ctx

    def start_patches(self, patches):
        for p in patches:
            p.start()

    def stop_patches(self, patches):
        for p in patches:
            p.stop()

    def test_check_for_setup_error(self):

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None

        jdssd = iscsi.JovianISCSIDriver()
        jdssd.configuration = cfg

        jdssd.ra = mock.Mock()

        # No IP
        jdssd.ra.is_pool_exists.return_value = True
        jdssd.jovian_hosts = []
        jdssd.block_size = ['64K']

        self.assertRaises(exception.VolumeDriverException,
                          jdssd.check_for_setup_error)

        # No pool detected
        jdssd.ra.is_pool_exists.return_value = False
        jdssd.jovian_hosts = ['192.168.0.2']
        jdssd.block_size = ['64K']

        self.assertRaises(exception.VolumeDriverException,
                          jdssd.check_for_setup_error)
        # Bad block size
        jdssd.ra.is_pool_exists.return_value = True
        jdssd.jovian_hosts = ['192.168.0.2', '192.168.0.3']
        jdssd.block_size = ['61K']

        self.assertRaises(exception.InvalidConfigurationValue,
                          jdssd.check_for_setup_error)

    def test_get_provider_location(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_1
        patches = [mock.patch.object(
            jdssd.ra,
            "get_active_host",
            return_value=host)]
        out = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        self.start_patches(patches)
        self.assertEqual(out, jdssd._get_provider_location(UUID_1))
        self.stop_patches(patches)

    def test_create_volume(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.size = 1
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]

        target_name = CONFIG_OK["target_prefix"] + UUID_1

        jdssd.ra.create_lun.return_value = None
        jdssd.ra.get_active_host.return_value = host

        ret = jdssd.create_volume(vol)

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        self.assertEqual(location, ret['provider_location'])
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))

    def test_create_volume_small_block(self):

        jdssd, ctx = self.get_driver(CONFIG_BLOCK_SIZE)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.size = 1
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]

        target_name = CONFIG_OK["target_prefix"] + UUID_1
        vname = jcom.vname(UUID_1)

        jdssd.ra.create_lun.return_value = None
        jdssd.ra.get_active_host.return_value = host

        ret = jdssd.create_volume(vol)

        jdssd.ra.create_lun.assert_called_once_with(
            vname, o_units.Gi, sparse=False, block_size="64K")

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        self.assertEqual(location, ret['provider_location'])
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))

    def test_hide_object(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        jdssd.ra.modify_lun.return_value = None
        jdssd._hide_object(vname)

        hidden_volume = {"name": jcom.hidden(UUID_1)}
        jdssd.ra.modify_lun.assert_called_once_with(vname, hidden_volume)

    def test_clean_garbage_snapshots(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        o_vname = jcom.vname(UUID_1)
        o_snaps = SNAPSHOTS_GARBAGE.copy()

        jdssd.ra.delete_snapshot.return_value = None
        jdssd._clean_garbage_snapshots(o_vname, o_snaps)
        jdssd.ra.delete_snapshot.assert_called_once_with(
            o_vname,
            SNAPSHOTS_GARBAGE[1]["name"])
        # Test exception handling
        for exc in get_jdss_exceptions():
            o_snaps = SNAPSHOTS_GARBAGE.copy()
            jdssd.ra.delete_snapshot.side_effect = exc
            try:
                jdssd._clean_garbage_snapshots(o_vname, o_snaps)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

    def test_cascade_volume_delete_snapshots(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        o_vname = jcom.vname(UUID_1)

        # Volume with 3 snapshots and no descendants
        # We should delete snapshots and then cal for volume deletion
        o_snaps = SNAPSHOTS_CASCADE_1.copy()

        jdssd.ra.modify_lun.return_value = None
        jdssd.ra.delete_snapshot.return_value = None
        jdssd.ra.get_snapshots.side_effect = [
            SNAPSHOTS_EMPTY,
            SNAPSHOTS_EMPTY,
            SNAPSHOTS_EMPTY]

        with mock.patch.object(jdssd, "_gc_delete", return_value=None) as gc:
            jdssd._cascade_volume_delete(o_vname, o_snaps)
            gc.assert_called_once_with(o_vname)
        delete_snapshot_expected = [
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_1[0]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True),
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_1[1]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True),
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_1[2]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True)]
        jdssd.ra.delete_snapshot.assert_has_calls(delete_snapshot_expected)

    def test_cascade_volume_delete_with_clone(self):
        # Volume with 2 snapshots and 1 clone
        # We should delete snapshots and then cal for volume hiding
        jdssd, ctx = self.get_driver(CONFIG_OK)
        o_vname = jcom.vname(UUID_1)
        o_snaps = SNAPSHOTS_CASCADE_2.copy()

        jdssd.ra.modify_lun.return_value = None
        jdssd.ra.delete_snapshot.return_value = None
        jdssd.ra.get_snapshots.side_effect = [
            SNAPSHOTS_EMPTY,
            SNAPSHOTS_EMPTY]

        patches = [mock.patch.object(jdssd, "_gc_delete"),
                   mock.patch.object(jdssd, "_hide_object")]

        self.start_patches(patches)

        jdssd._cascade_volume_delete(o_vname, o_snaps)

        jdssd._hide_object.assert_called_once_with(o_vname)
        jdssd._gc_delete.assert_not_called()

        self.stop_patches(patches)

        delete_snapshot_expected = [
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_2[0]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True),
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_2[2]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True)]
        jdssd.ra.delete_snapshot.assert_has_calls(delete_snapshot_expected)

    def test_cascade_volume_delete_snapshot_clone(self):
        # Volume with 3 snapshots and 1 clone of a snapshots
        # We should delete childless snapshots
        #           and then cal for volume deletion

        jdssd, ctx = self.get_driver(CONFIG_OK)
        o_vname = jcom.vname(UUID_1)
        o_snaps = SNAPSHOTS_CASCADE_1.copy()

        jdssd.ra.modify_lun.return_value = None
        jdssd.ra.delete_snapshot.return_value = None
        jdssd.ra.get_snapshots.side_effect = [
            SNAPSHOTS_EMPTY,
            SNAPSHOTS_CASCADE_3.copy(),
            SNAPSHOTS_EMPTY]
        get_snapshots = [
            mock.call(SNAPSHOTS_CASCADE_1[0]['name']),
            mock.call(SNAPSHOTS_CASCADE_1[1]['name']),
            mock.call(SNAPSHOTS_CASCADE_1[2]['name'])
        ]
        hide_object_expected = [
            mock.call(SNAPSHOTS_CASCADE_1[1]["name"]),
            mock.call(o_vname)]

        patches = [mock.patch.object(jdssd, "_gc_delete"),
                   mock.patch.object(jdssd, "_hide_object")]

        self.start_patches(patches)

        jdssd._cascade_volume_delete(o_vname, o_snaps)
        jdssd._hide_object.assert_has_calls(hide_object_expected)
        jdssd._gc_delete.assert_not_called()

        self.stop_patches(patches)

        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots)

        delete_snapshot_expected = [
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_2[0]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True),
            mock.call(o_vname,
                      SNAPSHOTS_CASCADE_2[2]["name"],
                      recursively_children=True,
                      recursively_dependents=True,
                      force_umount=True)]
        jdssd.ra.delete_snapshot.assert_has_calls(delete_snapshot_expected)

    def test_delete_volume_with_snapshots(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vname = jcom.vname(UUID_1)
        jdssd.ra.get_snapshots.side_effect = [SNAPSHOTS_CASCADE_1.copy()]

        patches = [mock.patch.object(jdssd, "_cascade_volume_delete"),
                   mock.patch.object(jdssd, "_gc_delete"),
                   mock.patch.object(jdssd, "_hide_object")]

        self.start_patches(patches)

        jdssd.delete_volume(vol, cascade=False)
        jdssd._gc_delete.assert_not_called()
        jdssd._cascade_volume_delete.assert_not_called()
        jdssd._hide_object.assert_called_once_with(vname)

        self.stop_patches(patches)

        jdssd.ra.get_snapshots.assert_called_once_with(vname)

    def test_delete_volume_without_snapshots(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vname = jcom.vname(UUID_1)
        jdssd.ra.get_snapshots.side_effect = [SNAPSHOTS_EMPTY.copy()]

        patches = [mock.patch.object(jdssd, "_cascade_volume_delete"),
                   mock.patch.object(jdssd, "_gc_delete"),
                   mock.patch.object(jdssd, "_hide_object")]

        self.start_patches(patches)

        jdssd.delete_volume(vol, cascade=False)
        jdssd._gc_delete.assert_called_once_with(vname)
        jdssd._cascade_volume_delete.assert_not_called()
        jdssd._hide_object.assert_not_called()

        self.stop_patches(patches)

        jdssd.ra.get_snapshots.assert_called_once_with(vname)

    def test_delete_volume_exceptions(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        patches = [mock.patch.object(jdssd, "_cascade_volume_delete"),
                   mock.patch.object(jdssd, "_gc_delete"),
                   mock.patch.object(jdssd, "_hide_object")]

        self.start_patches(patches)

        for exc in get_jdss_exceptions():
            jdssd.ra.get_snapshots.side_effect = exc
            try:
                jdssd.delete_volume(vol, cascade=False)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

        jdssd._gc_delete.assert_not_called()
        jdssd._cascade_volume_delete.assert_not_called()
        jdssd._hide_object.assert_not_called()

        self.stop_patches(patches)

    def test_gc_delete_not_clone(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        get_lun_resp = {'vscan': None,
                        'full_name': 'Pool-0/' + jcom.vname(UUID_1),
                        'userrefs': None,
                        'primarycache': 'all',
                        'logbias': 'latency',
                        'creation': '1591543140',
                        'sync': 'always',
                        'is_clone': False,
                        'dedup': 'off',
                        'sharenfs': None,
                        'receive_resume_token': None,
                        'volsize': '1073741824'}

        jdssd.ra.get_lun.return_value = get_lun_resp
        jdssd.ra.delete_lun.return_value = None

        patches = [mock.patch.object(jdssd, "_delete_back_recursively")]

        self.start_patches(patches)

        jdssd._gc_delete(jcom.vname(UUID_1))

        jdssd._delete_back_recursively.assert_not_called()
        jdssd.ra.delete_lun.assert_called_once_with(jcom.vname(UUID_1),
                                                    force_umount=True)

        self.stop_patches(patches)

    def test_gc_delete_is_clone(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        origin = "Pool-0/{vorig}@{sorig}".format(vorig=jcom.vname(UUID_1),
                                                 sorig=jcom.vname(UUID_2))

        get_lun_resp = {'origin': origin,
                        'vscan': None,
                        'full_name': 'Pool-0/' + jcom.vname(UUID_2),
                        'userrefs': None,
                        'primarycache': 'all',
                        'logbias': 'latency',
                        'creation': '1591543140',
                        'sync': 'always',
                        'is_clone': True,
                        'dedup': 'off',
                        'sharenfs': None,
                        'receive_resume_token': None,
                        'volsize': '1073741824'}

        jdssd.ra.get_lun.return_value = get_lun_resp
        jdssd.ra.delete_lun.return_value = None

        patches = [mock.patch.object(jdssd, "_delete_back_recursively")]

        self.start_patches(patches)

        jdssd._gc_delete(jcom.vname(UUID_1))

        jdssd._delete_back_recursively.assert_called_once_with(
            jcom.vname(UUID_1), jcom.vname(UUID_2))
        jdssd.ra.delete_lun.assert_not_called()

        self.stop_patches(patches)

    def test_delete_back_recursively_res_active(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        opvname = jcom.vname(UUID_1)
        opsname = jcom.sname(UUID_2)

        jdssd._delete_back_recursively(opvname, opsname)

        jdssd.ra.delete_snapshot.assert_called_once_with(
            opvname,
            opsname,
            recursively_children=True,
            recursively_dependents=True,
            force_umount=True)

    def test_delete_back_recursively_hidden_have_snapshots(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        opvname = jcom.hidden(UUID_1)
        opsname = jcom.sname(UUID_2)

        jdssd.ra.get_snapshots.return_value = SNAPSHOTS_RECURSIVE_1.copy()

        jdssd._delete_back_recursively(opvname, opsname)

        jdssd.ra.delete_snapshot.assert_called_once_with(
            opvname,
            opsname,
            recursively_children=True,
            recursively_dependents=True,
            force_umount=True)

    def test_delete_back_recursively_single_snapshot(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        opvname = jcom.hidden(UUID_2)
        opsname = jcom.sname(UUID_3)

        jdssd.ra.get_snapshots.side_effect = [
            SNAPSHOTS_RECURSIVE_CHAIN_1.copy(),
            SNAPSHOTS_RECURSIVE_CHAIN_2.copy()]

        origin = "Pool-0/{vorig}@{sorig}".format(vorig=jcom.vname(UUID_1),
                                                 sorig=jcom.vname(UUID_2))
        get_lun_resp = {'origin': origin,
                        'vscan': None,
                        'full_name': 'Pool-0/' + jcom.hidden(UUID_2),
                        'userrefs': None,
                        'primarycache': 'all',
                        'logbias': 'latency',
                        'creation': '1591543140',
                        'sync': 'always',
                        'is_clone': True,
                        'dedup': 'off',
                        'sharenfs': None,
                        'receive_resume_token': None,
                        'volsize': '1073741824'}
        jdssd.ra.get_lun.return_value = get_lun_resp
        jdssd._delete_back_recursively(opvname, opsname)

        jdssd.ra.delete_snapshot.assert_called_once_with(
            jcom.vname(UUID_1),
            jcom.vname(UUID_2),
            recursively_children=True,
            recursively_dependents=True,
            force_umount=True)

        get_snapshots_expected = [mock.call(opvname)]
        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots_expected)

    def test_extend_volume(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        jdssd.ra.extend_lun.return_value = None
        jdssd.extend_volume(vol, 2)

        jdssd.ra.extend_lun.assert_called_once_with(
            jcom.vname(UUID_1),
            2147483648)

    def test_extend_volume_exceptions(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        for exc in get_jdss_exceptions():
            try:
                jdssd.extend_volume(vol, 2)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

    def test_revert_to_snapshot(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        snap = fake_snapshot.fake_snapshot_obj(ctx)
        snap.id = UUID_2

        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        get_lun_resp_1 = {'vscan': None,
                          'full_name': 'Pool-0/' + UUID_1,
                          'userrefs': None,
                          'primarycache': 'all',
                          'logbias': 'latency',
                          'creation': '1591543140',
                          'sync': 'always',
                          'is_clone': False,
                          'dedup': 'off',
                          'sharenfs': None,
                          'receive_resume_token': None,
                          'volsize': '2147483648'}

        get_lun_resp_2 = {'vscan': None,
                          'full_name': 'Pool-0/' + UUID_1,
                          'userrefs': None,
                          'primarycache': 'all',
                          'logbias': 'latency',
                          'creation': '1591543140',
                          'sync': 'always',
                          'is_clone': False,
                          'dedup': 'off',
                          'sharenfs': None,
                          'receive_resume_token': None,
                          'volsize': '1073741824'}

        jdssd.ra.get_lun.side_effect = [get_lun_resp_1, get_lun_resp_2]

        get_lun_expected = [mock.call(vname), mock.call(vname)]

        jdssd.revert_to_snapshot(ctx, vol, snap)

        jdssd.ra.get_lun.assert_has_calls(get_lun_expected)

        jdssd.ra.rollback_volume_to_snapshot.assert_called_once_with(vname,
                                                                     sname)
        jdssd.ra.extend_lun(vname, '2147483648')

    def test_revert_to_snapshot_exception(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        snap = fake_snapshot.fake_snapshot_obj(ctx)
        snap.id = UUID_2

        vname = jcom.vname(UUID_1)

        get_lun_resp_no_size = {'vscan': None,
                                'full_name': 'Pool-0/' + vname,
                                'userrefs': None,
                                'primarycache': 'all',
                                'logbias': 'latency',
                                'creation': '1591543140',
                                'sync': 'always',
                                'is_clone': False,
                                'dedup': 'off',
                                'sharenfs': None,
                                'receive_resume_token': None,
                                'volsize': None}

        get_lun_resp_1 = {'vscan': None,
                          'full_name': 'Pool-0/' + vname,
                          'userrefs': None,
                          'primarycache': 'all',
                          'logbias': 'latency',
                          'creation': '1591543140',
                          'sync': 'always',
                          'is_clone': False,
                          'dedup': 'off',
                          'sharenfs': None,
                          'receive_resume_token': None,
                          'volsize': '2147483648'}

        get_lun_resp_2 = {'vscan': None,
                          'full_name': 'Pool-0/' + vname,
                          'userrefs': None,
                          'primarycache': 'all',
                          'logbias': 'latency',
                          'creation': '1591543140',
                          'sync': 'always',
                          'is_clone': False,
                          'dedup': 'off',
                          'sharenfs': None,
                          'receive_resume_token': None,
                          'volsize': '1073741824'}

        jdssd.ra.get_lun.side_effect = [get_lun_resp_no_size, get_lun_resp_2]

        self.assertRaises(exception.VolumeDriverException,
                          jdssd.revert_to_snapshot,
                          ctx,
                          vol,
                          snap)

        jdssd.ra.get_lun.side_effect = [get_lun_resp_1, get_lun_resp_2]

        jdssd.ra.rollback_volume_to_snapshot.side_effect = [
            jexc.JDSSResourceNotFoundException(res=vname)]
        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd.revert_to_snapshot,
                          ctx,
                          vol,
                          snap)

        jdssd.ra.get_lun.side_effect = [get_lun_resp_1,
                                        jexc.JDSSException("some_error")]

        jdssd.ra.rollback_volume_to_snapshot.side_effect = [
            jexc.JDSSResourceNotFoundException(res=vname)]
        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd.revert_to_snapshot,
                          ctx,
                          vol,
                          snap)

    def test_clone_object(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        origin = jcom.vname(UUID_1)
        clone = jcom.vname(UUID_2)

        jdssd.ra.create_snapshot.return_value = None
        jdssd.ra.create_volume_from_snapshot.return_value = None

        jdssd._clone_object(origin, clone)
        jdssd.ra.create_snapshot.assert_called_once_with(origin, clone)
        jdssd.ra.create_volume_from_snapshot.assert_called_once_with(
            clone,
            clone,
            origin,
            sparse=False)

    def test_clone_object_dne(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)
        calls = []
        origin = jcom.vname(UUID_1)
        clone = jcom.vname(UUID_2)
        calls.append(mock.call(origin, clone))

        jdssd.ra.create_snapshot.side_effect = (
            jexc.JDSSResourceNotFoundException(res=origin))

        self.assertRaises(exception.VolumeNotFound,
                          jdssd._clone_object, origin, clone)

        origin = jcom.sname(UUID_1)
        calls.append(mock.call(origin, clone))

        self.assertRaises(exception.SnapshotNotFound,
                          jdssd._clone_object, origin, clone)
        jdssd.ra.create_snapshot.assert_has_calls(calls)

    def test_clone_object_exists(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        origin = jcom.vname(UUID_1)
        clone = jcom.vname(UUID_2)

        jdssd.ra.create_snapshot.side_effect = (
            jexc.JDSSSnapshotExistsException(snapshot=clone))

        jdssd.ra.delete_snapshot.side_effect = (
            jexc.JDSSSnapshotIsBusyException(snapshot=clone))

        self.assertRaises(exception.Duplicate,
                          jdssd._clone_object, origin, clone)
        jdssd.ra.delete_snapshot.assert_called_once_with(origin, clone)
        jdssd.ra.create_snapshot.assert_called_once_with(origin, clone)

    def test_clone_object_volume_exists(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        origin = jcom.vname(UUID_1)
        clone = jcom.vname(UUID_2)

        jdssd.ra.create_snapshot.return_value = None
        jdssd.ra.create_volume_from_snapshot.side_effect = (
            jexc.JDSSVolumeExistsException(volume=clone))

        self.assertRaises(exception.Duplicate,
                          jdssd._clone_object, origin, clone)
        jdssd.ra.create_snapshot.assert_called_once_with(origin, clone)
        jdssd.ra.create_volume_from_snapshot.assert_called_once_with(
            clone,
            clone,
            origin,
            sparse=CONFIG_OK['san_thin_provision'])

    def test_create_cloned_volume(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        origin_vname = jcom.vname(UUID_1)
        clone_vname = jcom.vname(UUID_2)

        orig_vol = fake_volume.fake_volume_obj(ctx)
        orig_vol.id = UUID_1
        orig_vol.size = 1

        clone_vol = fake_volume.fake_volume_obj(ctx)
        clone_vol.id = UUID_2
        clone_vol.size = 1

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_2

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])

        patches = [
            mock.patch.object(jdssd, "_clone_object", return_value=None),
            mock.patch.object(jdssd, "extend_volume", return_value=None),
            mock.patch.object(
                jdssd,
                "_get_provider_location",
                return_value=location),
            mock.patch.object(
                jdssd,
                "_get_provider_auth",
                return_value=cred_format)]

        jdssd.ra.get_lun.return_value = {
            'vscan': None,
            'full_name': 'Pool-0/' + UUID_2,
            'userrefs': None,
            'primarycache': 'all',
            'logbias': 'latency',
            'creation': '1591543140',
            'sync': 'always',
            'is_clone': False,
            'dedup': 'off',
            'sharenfs': None,
            'receive_resume_token': None,
            'volsize': '1073741824'}

        self.start_patches(patches)

        ret = jdssd.create_cloned_volume(clone_vol, orig_vol)

        jdssd.extend_volume.assert_not_called()
        jdssd._clone_object.assert_called_once_with(origin_vname, clone_vname)
        self.stop_patches(patches)

        jdssd.ra.get_lun.assert_called_once_with(jcom.vname(clone_vol.id))
        self.assertEqual(location, ret['provider_location'])
        self.assertEqual(cred_format, ret['provider_auth'])

    def test_create_volume_from_snapshot(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        origin_sname = jcom.sname(UUID_1)

        clone_vname = jcom.vname(UUID_2)

        orig_snap = fake_snapshot.fake_snapshot_obj(ctx)
        orig_snap.id = UUID_1

        clone_vol = fake_volume.fake_volume_obj(ctx)
        clone_vol.id = UUID_2
        clone_vol.size = 1

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_2

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])

        patches = [
            mock.patch.object(jdssd, "_clone_object", return_value=None),
            mock.patch.object(jdssd, "extend_volume", return_value=None),
            mock.patch.object(
                jdssd,
                "_get_provider_location",
                return_value=location),
            mock.patch.object(
                jdssd,
                "_get_provider_auth",
                return_value=cred_format)]

        jdssd.ra.get_lun.return_value = {
            'vscan': None,
            'full_name': 'Pool-0/' + UUID_2,
            'userrefs': None,
            'primarycache': 'all',
            'logbias': 'latency',
            'creation': '1591543140',
            'sync': 'always',
            'is_clone': False,
            'dedup': 'off',
            'sharenfs': None,
            'receive_resume_token': None,
            'volsize': '1073741824'}

        self.start_patches(patches)

        ret = jdssd.create_volume_from_snapshot(clone_vol, orig_snap)

        jdssd.extend_volume.assert_not_called()
        jdssd._clone_object.assert_called_once_with(origin_sname, clone_vname)
        self.stop_patches(patches)

        jdssd.ra.get_lun.assert_called_once_with(jcom.vname(clone_vol.id))
        self.assertEqual(location, ret['provider_location'])
        self.assertEqual(cred_format, ret['provider_auth'])

    def test_create_volume_from_snapshot_extend(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        origin_sname = jcom.sname(UUID_1)

        clone_vname = jcom.vname(UUID_2)

        orig_snap = fake_snapshot.fake_snapshot_obj(ctx)
        orig_snap.id = UUID_1

        clone_vol = fake_volume.fake_volume_obj(ctx)
        clone_vol.id = UUID_2
        clone_vol.size = 2

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_2

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])

        patches = [
            mock.patch.object(jdssd, "_clone_object", return_value=None),
            mock.patch.object(jdssd, "extend_volume", return_value=None),
            mock.patch.object(
                jdssd,
                "_get_provider_location",
                return_value=location),
            mock.patch.object(
                jdssd,
                "_get_provider_auth",
                return_value=cred_format)]

        jdssd.ra.get_lun.return_value = {
            'vscan': None,
            'full_name': 'Pool-0/' + UUID_2,
            'userrefs': None,
            'primarycache': 'all',
            'logbias': 'latency',
            'creation': '1591543140',
            'sync': 'always',
            'is_clone': False,
            'dedup': 'off',
            'sharenfs': None,
            'receive_resume_token': None,
            'volsize': '1073741824'}

        self.start_patches(patches)

        ret = jdssd.create_volume_from_snapshot(clone_vol, orig_snap)

        jdssd.extend_volume.assert_called_once_with(clone_vol, clone_vol.size)
        jdssd._clone_object.assert_called_once_with(origin_sname, clone_vname)
        self.stop_patches(patches)

        jdssd.ra.get_lun.assert_called_once_with(jcom.vname(clone_vol.id))
        self.assertEqual(location, ret['provider_location'])
        self.assertEqual(cred_format, ret['provider_auth'])

    def test_create_snapshot(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        vname = jcom.vname(UUID_1)

        sname = jcom.sname(UUID_2)

        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_2)
        snap.volume_id = UUID_1

        patches = [
            mock.patch.object(jdssd, "_clone_object", return_value=None)]

        self.start_patches(patches)

        jdssd.create_snapshot(snap)

        jdssd._clone_object.assert_called_once_with(vname, sname)
        self.stop_patches(patches)

    def test_delete_snapshot_no_child(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        sname = jcom.sname(UUID_2)

        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_2)

        jdssd.ra.get_snapshots.return_value = SNAPSHOTS_EMPTY
        patches = [
            mock.patch.object(
                jdssd,
                "_clean_garbage_snapshots",
                return_value=SNAPSHOTS_EMPTY),
            mock.patch.object(jdssd, "_clone_object", return_value=None),
            mock.patch.object(jdssd, "_hide_object", return_value=None),
            mock.patch.object(jdssd, "_gc_delete", return_value=None)]

        self.start_patches(patches)

        jdssd.create_snapshot(snap)

        jdssd.delete_snapshot(snap)
        jdssd._gc_delete.assert_called_once_with(sname)
        jdssd._hide_object.assert_not_called()
        jdssd._clean_garbage_snapshots.assert_called_once_with(
            sname,
            SNAPSHOTS_EMPTY)
        self.stop_patches(patches)

    def test_delete_snapshot_has_clone(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        sname = jcom.sname(UUID_2)

        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_2)

        jdssd.ra.get_snapshots.return_value = SNAPSHOTS_EMPTY
        patches = [
            mock.patch.object(
                jdssd,
                "_clean_garbage_snapshots",
                return_value=SNAPSHOTS_CLONE),
            mock.patch.object(jdssd, "_clone_object", return_value=None),
            mock.patch.object(jdssd, "_hide_object", return_value=None),
            mock.patch.object(jdssd, "_gc_delete", return_value=None)]

        self.start_patches(patches)

        jdssd.create_snapshot(snap)

        jdssd.delete_snapshot(snap)
        jdssd._gc_delete.assert_not_called()
        jdssd._hide_object.assert_called_once_with(sname)
        jdssd._clean_garbage_snapshots.assert_called_once_with(
            sname,
            SNAPSHOTS_EMPTY)
        self.stop_patches(patches)

    def test_local_path(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_1)

        self.assertRaises(NotImplementedError, jdssd.local_path, vol)

    def test_get_provider_auth(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        auth = jdssd._get_provider_auth()
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, auth))

    def test_get_provider_auth_long(self):
        long_pass_config = CONFIG_OK.copy()
        long_pass_config['chap_password_len'] = 16
        jdssd, ctx = self.get_driver(long_pass_config)

        auth = jdssd._get_provider_auth()
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=16)
        self.assertIsNotNone(re.match(cred_format, auth))

    def test_create_export(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        patches = [
            mock.patch.object(
                jdssd,
                "_ensure_target_volume",
                return_value=None),
            mock.patch.object(
                jdssd,
                "_get_provider_location",
                return_value='provider_location')]

        self.start_patches(patches)

        ret = jdssd.create_export(ctx, vol, "connector")
        jdssd._ensure_target_volume.assert_called_once_with(vol)

        self.stop_patches(patches)

        self.assertEqual('provider_location', ret["provider_location"])

    def test_ensure_export(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        patches = [
            mock.patch.object(
                jdssd,
                "_ensure_target_volume",
                return_value=None)]

        self.start_patches(patches)

        jdssd.ensure_export(ctx, vol)
        jdssd._ensure_target_volume.assert_called_once_with(vol)

        self.stop_patches(patches)

    def test_remove_export(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx, id=UUID_1)

        patches = [
            mock.patch.object(
                jdssd,
                "_remove_target_volume",
                return_value=None)]

        self.start_patches(patches)

        jdssd.remove_export(ctx, vol)
        jdssd._remove_target_volume.assert_called_once_with(vol)

        self.stop_patches(patches)

    def test_update_volume_stats(self):
        jdssd, ctx = self.get_driver(CONFIG_BACKEND_NAME)

        location_info = 'JovianISCSIDriver:192.168.0.2:Pool-0'
        correct_out = {
            'vendor_name': 'Open-E',
            'driver_version': "1.0.2",
            'storage_protocol': 'iSCSI',
            'total_capacity_gb': 100,
            'free_capacity_gb': 50,
            'reserved_percentage': 10,
            'volume_backend_name': CONFIG_BACKEND_NAME['volume_backend_name'],
            'QoS_support': False,
            'location_info': location_info,
            'multiattach': True
        }
        jdssd.ra.get_pool_stats.return_value = {
            'size': 100 * o_units.Gi,
            'available': 50 * o_units.Gi}
        jdssd.ra.get_active_host.return_value = CONFIG_OK['san_hosts']
        jdssd._update_volume_stats()

        self.assertEqual(correct_out, jdssd._stats)

    def test_create_target(self):

        jdssd, ctx = self.get_driver(CONFIG_BACKEND_NAME)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        jdssd.ra.create_target.return_value = None
        jdssd._create_target(target_name, use_chap=True)

        jdssd.ra.create_target.assert_called_once_with(
            target_name, use_chap=True)

        jdssd.ra.create_target.side_effect = jexc.JDSSResourceExistsException(
            res=target_name)

        self.assertRaises(exception.Duplicate,
                          jdssd._create_target,
                          target_name,
                          use_chap=True)

    def test_attach_target_volume(self):

        jdssd, ctx = self.get_driver(CONFIG_BACKEND_NAME)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)

        jdssd.ra.attach_target_vol.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._attach_target_volume(target_name, vname)

        jdssd.ra.attach_target_vol.assert_called_once_with(
            target_name, vname)
        jdssd.ra.delete_target.assert_not_called()

        ex = jexc.JDSSResourceExistsException(res=target_name)
        jdssd.ra.attach_target_vol.side_effect = ex

        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd._attach_target_volume,
                          target_name,
                          vname)
        jdssd.ra.delete_target.assert_called_once_with(target_name)

    def test_set_target_credentials(self):
        jdssd, ctx = self.get_driver(CONFIG_BACKEND_NAME)

        target_name = CONFIG_BACKEND_NAME['target_prefix'] + UUID_1
        cred = {'name': 'user_name', 'password': '123456789012'}

        jdssd.ra.create_target_user.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._set_target_credentials(target_name, cred)

        jdssd.ra.create_target_user.assert_called_once_with(
            target_name, cred)
        jdssd.ra.delete_target.assert_not_called()

        ex = jexc.JDSSResourceExistsException(res=target_name)
        jdssd.ra.create_target_user.side_effect = ex

        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd._set_target_credentials,
                          target_name,
                          cred)
        jdssd.ra.delete_target.assert_called_once_with(target_name)

    def test_create_target_volume(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name', 'password': '123456789012'}

        patches = [
            mock.patch.object(jdssd, "_create_target"),
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials")]

        self.start_patches(patches)
        jdssd._create_target_volume(vol)
        jdssd._create_target.assert_called_once_with(target_name, True)
        jdssd._attach_target_volume.assert_called_once_with(
            target_name, jcom.vname(UUID_1))
        jdssd._set_target_credentials.assert_called_once_with(
            target_name, cred)
        self.stop_patches(patches)

    def test_ensure_target_volume(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vname = jcom.vname(UUID_1)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_create_target"),
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = True
        jdssd.ra.is_target_lun.return_value = True
        jdssd.ra.get_target_user.return_value = [cred]

        self.start_patches(patches)

        jdssd._ensure_target_volume(vol)

        jdssd.ra.is_target.assert_called_once_with(target_name)

        jdssd.ra.is_target_lun.assert_called_once_with(target_name, vname)

        jdssd.ra.get_target_user.assert_called_once_with(target_name)

        jdssd.ra.delete_target_user.assert_not_called()
        jdssd._set_target_credentials.assert_not_called()
        self.stop_patches(patches)

    def test_ensure_target_volume_not_attached(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vname = jcom.vname(UUID_1)
        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_create_target"),
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = True
        jdssd.ra.is_target_lun.return_value = False
        jdssd.ra.get_target_user.return_value = [cred]

        self.start_patches(patches)

        jdssd._ensure_target_volume(vol)

        jdssd.ra.is_target.assert_called_once_with(target_name)
        jdssd.ra.is_target_lun.assert_called_once_with(target_name, vname)

        jdssd._attach_target_volume.assert_called_once_with(
            target_name, vname)
        jdssd.ra.get_target_user.assert_called_once_with(target_name)

        jdssd.ra.delete_target_user.assert_not_called()
        jdssd._set_target_credentials.assert_not_called()
        self.stop_patches(patches)

    def test_ensure_target_volume_no_target(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_create_target_volume"),
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = False
        jdssd.ra.get_target_user.return_value = cred['name']

        self.start_patches(patches)

        jdssd._ensure_target_volume(vol)

        jdssd.ra.is_target.assert_called_once_with(target_name)
        jdssd._create_target_volume.assert_called_once_with(vol)

        jdssd.ra.is_target_lun.assert_not_called()
        self.stop_patches(patches)

    def test_remove_target_volume(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._remove_target_volume(vol)

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_remove_target_volume_no_target(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.detach_target_vol.side_effect = (
            jexc.JDSSResourceNotFoundException(res=target_name))
        jdssd.ra.delete_target.return_value = None

        jdssd._remove_target_volume(vol)

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_remove_target_volume_fail_to_detach(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.side_effect = (
            jexc.JDSSRESTException(reason='running test', request='test'))
        jdssd.ra.delete_target.return_value = None

        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd._remove_target_volume, vol)

        jdssd.ra.detach_target_vol.assert_called_once_with(
            target_name, jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_not_called()

    def test_remove_target_volume_fail_to_delete(self):

        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.delete_target.side_effect = (
            jexc.JDSSRESTException(reason='running test', request='test'))

        self.assertRaises(exception.VolumeBackendAPIException,
                          jdssd._remove_target_volume, vol)

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_get_iscsi_properties(self):
        jdssd, ctx = self.get_driver(CONFIG_OK)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        connector = {'multipath': True}

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        ret = jdssd._get_iscsi_properties(vol, connector)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqns': [target_name],
                    'target_lun': 0,
                    'target_luns': [0],
                    'target_portals': ['192.168.0.2:3260']}
        self.assertEqual(expected, ret)

    def test_get_iscsi_properties_multipath(self):
        jdssd, ctx = self.get_driver(CONFIG_MULTI_HOST)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        connector = {'multipath': True}

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        ret = jdssd._get_iscsi_properties(vol, connector)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqns': [target_name, target_name],
                    'target_lun': 0,
                    'target_luns': [0, 0],
                    'target_portals': ['192.168.0.2:3260', '192.168.0.3:3260']}
        self.assertEqual(expected, ret)

    def test_initialize_connection(self):

        jdssd, ctx = self.get_driver(CONFIG_MULTI_HOST)

        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.provider_auth = 'chap user_name 123456789012'

        connector = {'multipath': True, 'ip': '172.16.0.2'}

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqns': [target_name, target_name],
                      'target_lun': 0,
                      'target_luns': [0, 0],
                      'target_portals': ['192.168.0.2:3260',
                                         '192.168.0.3:3260']}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

        ret = jdssd.initialize_connection(vol, connector)

        self.assertEqual(con_info, ret)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import attachments
from cinder import exception
from cinder.policies import attachments as attachments_policies
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils
from cinder.volume import manager as volume_manager


@ddt.ddt
class AttachmentsPolicyTest(base.BasePolicyTest):
    authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)

        self.controller = attachments.AttachmentsController()
        self.manager = volume_manager.VolumeManager()
        self.manager.driver = mock.MagicMock()
        self.manager.driver.initialize_connection = mock.MagicMock()
        self.manager.driver.initialize_connection.side_effect = (
            self._initialize_connection)

        self.api_path = '/v3/%s/attachments' % (self.project_id)
        self.api_version = mv.NEW_ATTACH

    def _initialize_connection(self, volume, connector):
        return {'data': connector}

    def _create_attachment(self):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)
        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=vol_type.id,
                                          admin_metadata={
                                              'attached_mode': 'ro'
                                          },
                                          testcase_instance=self)
        volume = test_utils.attach_volume(self.project_member_context,
                                          volume.id,
                                          fake.INSTANCE_ID,
                                          'fake_host',
                                          'fake_mountpoint')
        return volume.volume_attachment[0].id

    @ddt.data(*base.all_users)
    def test_create_attachment_policy(self, user_id):
        volume = test_utils.create_volume(self.project_member_context,
                                          testcase_instance=self)
        rule_name = attachments_policies.CREATE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "attachment": {
                "instance_uuid": fake.INSTANCE_ID,
                "volume_uuid": volume.id,
            }
        }

        # Some context return HTTP 404 (rather than 403).
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_update_attachment_policy(self, user_id, mock_attachment_update):
        # Redirect the RPC call directly to the volume manager.
        def attachment_update(*args):
            return self.manager.attachment_update(*args)

        mock_attachment_update.side_effect = attachment_update

        rule_name = attachments_policies.UPDATE_POLICY
        attachment_id = self._create_attachment()
        url = '%s/%s' % (self.api_path, attachment_id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'PUT'
        body = {
            "attachment": {
                "connector": {
                    "initiator": "iqn.1993-08.org.debian: 01: cad181614cec",
                    "ip": "192.168.1.20",
                    "platform": "x86_64",
                    "host": "tempest-1",
                    "os_type": "linux2",
                    "multipath": False,
                    "mountpoint": "/dev/vdb",
                    "mode": "ro"
                }
            }
        }

        unauthorized_exceptions = []

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.update, req,
                                 id=attachment_id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_delete')
    def test_delete_attachment_policy(self, user_id, mock_attachment_delete):
        # Redirect the RPC call directly to the volume manager.
        def attachment_delete(*args):
            return self.manager.attachment_delete(*args)

        mock_attachment_delete.side_effect = attachment_delete

        rule_name = attachments_policies.DELETE_POLICY
        attachment_id = self._create_attachment()
        url = '%s/%s' % (self.api_path, attachment_id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'

        unauthorized_exceptions = []

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.delete, req,
                                 id=attachment_id)

    @ddt.data(*base.all_users)
    def test_complete_attachment_policy(self, user_id):
        rule_name = attachments_policies.COMPLETE_POLICY
        attachment_id = self._create_attachment()
        url = '%s/%s/action' % (self.api_path, attachment_id)
        req = fake_api.HTTPRequest.blank(url, version=mv.NEW_ATTACH_COMPLETION)
        req.method = 'POST'
        body = {
            "os-complete": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.complete, req,
                                 id=attachment_id, body=body)

    @ddt.data(*base.all_users)
    def test_multiattach_bootable_volume_policy(self, user_id):
        volume = test_utils.create_volume(self.project_member_context,
                                          multiattach=True,
                                          status='in-use',
                                          bootable=True,
                                          testcase_instance=self)
        rule_name = attachments_policies.MULTIATTACH_BOOTABLE_VOLUME_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "attachment": {
                "instance_uuid": fake.INSTANCE_ID,
                "volume_uuid": volume.id,
            }
        }

        # Relax the CREATE_POLICY in order to get past that check, which takes
        # place prior to checking the MULTIATTACH_BOOTABLE_VOLUME_POLICY.
        self.policy.set_rules({attachments_policies.CREATE_POLICY: ""},
                              overwrite=False)

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)


class AttachmentsPolicySecureRbacTest(AttachmentsPolicyTest):
    authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]

    unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.tests.functional.api_sample_tests import fakes
from cinder.tests.functional import api_samples_test_base as test_base


class VolumeActionsSampleJsonTest(test_base.VolumesSampleBase):
    sample_dir = "volume_actions"

    def setUp(self):
        super(VolumeActionsSampleJsonTest, self).setUp()
        self.response = self._create_volume()
        self.stub_out("cinder.volume.api.API.copy_volume_to_image",
                      fakes.stub_copy_volume_to_image)

    def test_volume_upload_image(self):
        res = jsonutils.loads(self.response.content)['volume']
        self._poll_volume_while(res['id'], ['creating'])
        response = self._do_post('volumes/%s/action' % res['id'],
                                 'volume-upload-to-image-request')
        self._verify_response('volume-upload-to-image-response',
                              {}, response, 202)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The limits V3 api."""

from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.v2 import limits as limits_v2
from cinder.api.views import limits as limits_views
from cinder import quota

QUOTAS = quota.QUOTAS


class LimitsController(limits_v2.LimitsController):
    """Controller for accessing limits in the OpenStack API."""

    def index(self, req):
        """Return all global and rate limit information."""
        context = req.environ['cinder.context']
        params = req.params.copy()
        req_version = req.api_version_request

        # TODO(wangxiyuan): Support "tenant_id" here to keep the backwards
        # compatibility. Remove it once we drop all support for "tenant".
        if (req_version.matches(None,
                                mv.get_prior_version(mv.LIMITS_ADMIN_FILTER))
                or not context.is_admin):
            params.pop('project_id', None)
            params.pop('tenant_id', None)
        project_id = params.get(
            'project_id', params.get('tenant_id', context.project_id))

        quotas = QUOTAS.get_project_quotas(context, project_id,
                                           usages=False)
        abs_limits = {k: v['limit'] for k, v in quotas.items()}
        rate_limits = req.environ.get("cinder.limits", [])

        builder = self._get_view_builder(req)
        return builder.build(rate_limits, abs_limits)

    def _get_view_builder(self, req):
        return limits_views.ViewBuilder()


def create_resource():
    return wsgi.Resource(LimitsController())
